import {
  getErrorMessageFromFramesStackItem,
  isPartialFrameStackItem
} from "./chunk-NKAZJQ23.js";

// src/ui/frame.base.tsx
import {
  createElement as reactCreateElement,
  useCallback,
  useEffect,
  useRef,
  useState
} from "react";
function defaultMessageHandler() {
}
function BaseFrameUI({
  frameState,
  components,
  theme,
  allowPartialFrame = false,
  enableImageDebugging = false,
  // eslint-disable-next-line no-console -- provide at least some feedback to the user
  onError = console.error,
  onMessage = defaultMessageHandler,
  createElement = reactCreateElement
}) {
  const [isImageLoading, setIsImageLoading] = useState(true);
  const { currentFrameStackItem } = frameState;
  const rootRef = useRef(null);
  const rootDimensionsRef = useRef();
  const onImageLoadEnd = useCallback(() => {
    setIsImageLoading(false);
  }, []);
  useEffect(() => {
    if (currentFrameStackItem?.status === "pending") {
      setIsImageLoading(true);
    }
  }, [currentFrameStackItem?.status]);
  const onErrorRef = useRef(onError);
  onErrorRef.current = onError;
  useEffect(() => {
    if (currentFrameStackItem?.status === "requestError") {
      onErrorRef.current(currentFrameStackItem.requestError);
    }
  }, [currentFrameStackItem]);
  const onMessageRef = useRef(onMessage);
  onMessageRef.current = onMessage;
  useEffect(() => {
    if (currentFrameStackItem?.status === "message") {
      onMessageRef.current({
        message: currentFrameStackItem.message,
        status: currentFrameStackItem.type === "info" ? "message" : "error"
      });
    }
  }, [currentFrameStackItem]);
  if (!frameState.homeframeUrl) {
    return components.Error(
      { message: "Missing frame url" },
      theme?.Error || {}
    );
  }
  if (!currentFrameStackItem) {
    return null;
  }
  let frameUiState;
  const previousFrame = frameState.framesStack[frameState.framesStack.length - 1]?.frameResult?.frame;
  switch (currentFrameStackItem.status) {
    case "requestError": {
      if ("sourceFrame" in currentFrameStackItem.request && currentFrameStackItem.request.sourceFrame) {
        frameUiState = {
          status: "complete",
          frame: currentFrameStackItem.request.sourceFrame,
          isImageLoading,
          id: currentFrameStackItem.timestamp.getTime(),
          frameState
        };
      } else {
        return components.Error(
          {
            message: "Failed to load frame",
            error: currentFrameStackItem.requestError
          },
          theme?.Error || {}
        );
      }
      break;
    }
    case "message":
      if (!currentFrameStackItem.request.sourceFrame) {
        return components.Message(
          {
            message: currentFrameStackItem.message,
            status: currentFrameStackItem.type === "error" ? "error" : "message"
          },
          theme?.Message ?? {}
        );
      }
      frameUiState = {
        status: "complete",
        frame: currentFrameStackItem.request.sourceFrame,
        isImageLoading,
        id: currentFrameStackItem.timestamp.getTime(),
        frameState
      };
      break;
    case "doneRedirect": {
      if (!currentFrameStackItem.request.sourceFrame) {
        frameUiState = {
          status: "loading",
          id: currentFrameStackItem.timestamp.getTime(),
          frameState
        };
      } else {
        frameUiState = {
          status: "complete",
          frame: currentFrameStackItem.request.sourceFrame,
          isImageLoading,
          id: currentFrameStackItem.timestamp.getTime(),
          frameState
        };
      }
      break;
    }
    case "done": {
      if (currentFrameStackItem.frameResult.status === "success") {
        frameUiState = {
          status: "complete",
          frame: currentFrameStackItem.frameResult.frame,
          debugImage: enableImageDebugging ? currentFrameStackItem.frameResult.framesDebugInfo?.image : void 0,
          isImageLoading,
          id: currentFrameStackItem.timestamp.getTime(),
          frameState
        };
      } else if (isPartialFrameStackItem(currentFrameStackItem) && allowPartialFrame) {
        frameUiState = {
          status: "partial",
          frame: currentFrameStackItem.frameResult.frame,
          debugImage: enableImageDebugging ? currentFrameStackItem.frameResult.framesDebugInfo?.image : void 0,
          isImageLoading,
          id: currentFrameStackItem.timestamp.getTime(),
          frameState
        };
      } else {
        return components.Error(
          { message: "Invalid frame" },
          theme?.Error || {}
        );
      }
      break;
    }
    case "pending": {
      frameUiState = {
        status: "loading",
        id: currentFrameStackItem.timestamp.getTime(),
        frameState
      };
      break;
    }
  }
  const isLoading = frameUiState.status === "loading" || frameUiState.isImageLoading;
  return components.Root(
    {
      createElement,
      frameState: frameUiState,
      dimensions: isLoading ? rootDimensionsRef.current ?? null : null,
      ref: rootRef,
      loadingScreen: isLoading ? components.LoadingScreen(
        {
          frameState: frameUiState,
          dimensions: rootDimensionsRef.current ?? null
        },
        theme?.LoadingScreen || {}
      ) : null,
      buttonsContainer: frameUiState.status === "loading" || !frameUiState.frame.buttons || frameUiState.frame.buttons.length === 0 ? null : components.ButtonsContainer(
        {
          frameState: frameUiState,
          buttons: frameUiState.frame.buttons.map(
            (frameButton, index) => components.Button(
              {
                frameState: frameUiState,
                frameButton,
                index,
                // @TODO provide previous frame to pending state so we can remove loading check and render some loading indicator?
                isDisabled: false,
                onPress() {
                  rootDimensionsRef.current = rootRef.current?.computeDimensions();
                  Promise.resolve(
                    frameState.onButtonPress(
                      // @todo change the type onButtonPress to accept partial frame as well because that can happen if partial frames are enabled
                      frameUiState.frame,
                      frameButton,
                      index
                    )
                  ).catch((error) => {
                    console.error(error);
                  });
                }
              },
              theme?.Button || {}
            )
          )
        },
        theme?.ButtonsContainer || {}
      ),
      imageContainer: components.ImageContainer(
        {
          frameState: frameUiState,
          aspectRatio: (frameUiState.status === "loading" ? previousFrame?.imageAspectRatio : frameUiState.frame.imageAspectRatio) ?? "1.91:1",
          image: components.Image(
            frameUiState.status === "loading" ? {
              status: "frame-loading",
              frameState: frameUiState,
              onImageLoadEnd,
              aspectRatio: previousFrame?.imageAspectRatio ?? "1.91:1"
            } : {
              status: "frame-loading-complete",
              frameState: frameUiState,
              src: frameUiState.debugImage ?? frameUiState.frame.image,
              aspectRatio: frameUiState.frame.imageAspectRatio ?? "1.91:1",
              onImageLoadEnd
            },
            theme?.Image || {}
          ),
          messageTooltip: currentFrameStackItem.status === "message" ? components.MessageTooltip(
            {
              frameState: frameUiState,
              status: currentFrameStackItem.type === "error" ? "error" : "message",
              message: getErrorMessageFromFramesStackItem(
                currentFrameStackItem
              )
            },
            theme?.MessageTooltip || {}
          ) : null
        },
        theme?.ImageContainer || {}
      ),
      textInputContainer: frameUiState.status === "loading" || typeof frameUiState.frame.inputText !== "string" ? null : components.TextInputContainer(
        {
          frame: frameUiState,
          textInput: components.TextInput(
            {
              // @TODO provide previous frame to pending state so we can remove loading check and render skeletons?
              isDisabled: false,
              frameState: frameUiState,
              placeholder: frameUiState.frame.inputText,
              onChange(text) {
                frameState.setInputText(text);
              },
              value: frameState.inputText
            },
            theme?.TextInput || {}
          )
        },
        theme?.TextInputContainer || {}
      )
    },
    theme?.Root || {}
  );
}

export {
  BaseFrameUI
};
