import {
  isParseResult
} from "./chunk-B237QKYM.js";

// src/use-fetch-frame.ts
import { hexToBytes } from "viem";
var UnexpectedCastActionResponseError = class extends Error {
  constructor() {
    super("Unexpected cast action response from the server");
  }
};
var UnexpectedComposerActionResponseError = class extends Error {
  constructor() {
    super("Unexpected composer action response from the server");
  }
};
function isErrorMessageResponse(response) {
  return typeof response === "object" && response !== null && "message" in response && typeof response.message === "string";
}
function isComposerFormActionResponse(response) {
  return typeof response === "object" && response !== null && "type" in response && response.type === "form";
}
function isCastActionFrameResponse(response) {
  return typeof response === "object" && response !== null && "type" in response && response.type === "frame";
}
function isCastMessageResponse(response) {
  return typeof response === "object" && response !== null && "message" in response && typeof response.message === "string";
}
function defaultErrorHandler(error) {
  console.error(error);
}
function useFetchFrame({
  stackAPI,
  stackDispatch,
  specification,
  frameActionProxy,
  frameGetProxy,
  extraButtonRequestPayload,
  signFrameAction,
  onTransaction,
  transactionDataSuffix,
  onSignature,
  onError = defaultErrorHandler,
  fetchFn,
  onRedirect,
  onComposerFormAction
}) {
  async function handleFailedResponse({
    response,
    endTime,
    frameStackPendingItem
  }) {
    if (response.ok) {
      throw new TypeError(
        "handleFailedResponse called with a successful response"
      );
    }
    const responseBody = await getResponseBody(response);
    if (response.status >= 400 && response.status < 500) {
      if (frameStackPendingItem.method === "POST" && isErrorMessageResponse(responseBody)) {
        stackAPI.markAsDoneWithErrorMessage({
          pendingItem: frameStackPendingItem,
          endTime,
          response,
          responseData: responseBody
        });
        onError(new Error(responseBody.message));
        return;
      }
    }
    const requestError = new Error(
      `The server returned an error but it does not contain message property. Status code: ${response.status}`
    );
    stackAPI.markAsFailed({
      endTime,
      pendingItem: frameStackPendingItem,
      requestError,
      responseStatus: response.status,
      response,
      responseBody
    });
    onError(requestError);
  }
  async function fetchGETRequest(request, shouldClear) {
    if (shouldClear) {
      stackDispatch({ action: "CLEAR" });
    }
    const frameStackPendingItem = stackAPI.createGetPendingItem({ request });
    const response = await fetchProxied({
      proxyUrl: frameGetProxy,
      fetchFn,
      specification,
      url: request.url
    });
    const endTime = /* @__PURE__ */ new Date();
    if (response instanceof Response) {
      if (!response.ok) {
        await handleFailedResponse({
          response,
          endTime,
          frameStackPendingItem
        });
        return;
      }
      const frameResult = await response.clone().json();
      stackAPI.markAsDone({
        pendingItem: frameStackPendingItem,
        endTime,
        frameResult,
        response
      });
      return;
    }
    stackAPI.markAsFailed({
      pendingItem: frameStackPendingItem,
      endTime,
      requestError: response,
      response: null,
      responseBody: "none",
      responseStatus: 500
    });
    onError(response);
  }
  async function fetchPOSTRequest(request, options, shouldClear) {
    let pendingItem;
    if (shouldClear) {
      stackDispatch({ action: "CLEAR" });
    }
    const signedDataOrError = await signAndGetFrameActionBodyPayload({
      signerStateActionContext: request.signerStateActionContext,
      signFrameAction
    });
    if (signedDataOrError instanceof Error) {
      if (options) {
        stackAPI.markAsFailed({
          pendingItem: options.pendingFrameStackItem,
          endTime: /* @__PURE__ */ new Date(),
          requestError: signedDataOrError,
          response: null,
          // there is no response because didn't even got to request
          responseBody: "none",
          responseStatus: 500
        });
      }
      onError(signedDataOrError);
      return;
    }
    if (!options) {
      pendingItem = stackAPI.createPostPendingItem({
        action: signedDataOrError,
        request
      });
    } else {
      pendingItem = options.pendingFrameStackItem;
    }
    const response = await fetchProxied({
      proxyUrl: frameActionProxy,
      specification,
      fetchFn,
      frameAction: signedDataOrError,
      extraRequestPayload: extraButtonRequestPayload
    });
    const endTime = /* @__PURE__ */ new Date();
    async function handleRedirect(res, currentPendingItem) {
      try {
        let location = res.headers.get("location");
        if (!location) {
          const responseData = await res.clone().json();
          if (responseData && typeof responseData === "object" && "location" in responseData && typeof responseData.location === "string") {
            location = responseData.location;
          }
        }
        if (!location) {
          throw new Error(
            `Response data does not contain 'location' key and no 'location' header is found.`
          );
        }
        const locationUrl = new URL(location);
        onRedirect(locationUrl);
        stackAPI.markAsDoneWithRedirect({
          pendingItem: currentPendingItem,
          endTime,
          location,
          response: res.clone(),
          responseBody: await res.clone().text()
        });
      } catch (e) {
        const error = e instanceof Error ? e : new Error(
          "Response body must be a json with 'location' property or response 'Location' header must contain fully qualified URL."
        );
        stackAPI.markAsFailedWithRequestError({
          pendingItem: currentPendingItem,
          error,
          response: res,
          endTime,
          responseBody: await res.clone().text()
        });
        onError(error);
      }
    }
    if (response instanceof Response) {
      if (response.status === 302) {
        await handleRedirect(response, pendingItem);
        return;
      }
      if (!response.ok) {
        await handleFailedResponse({
          response,
          endTime,
          frameStackPendingItem: pendingItem
        });
        return;
      }
      const responseData = await response.clone().json();
      if (!isParseResult(responseData)) {
        const error = new Error(`The server returned an unexpected response.`);
        stackAPI.markAsFailed({
          endTime,
          pendingItem,
          requestError: error,
          response,
          responseBody: responseData,
          responseStatus: 500
        });
        onError(error);
        return;
      }
      stackAPI.markAsDone({
        endTime,
        frameResult: responseData,
        pendingItem,
        response
      });
      return;
    }
    stackAPI.markAsFailed({
      endTime,
      pendingItem,
      requestError: response,
      response: new Response(response.message, {
        status: 500,
        headers: { "Content-Type": "text/plain" }
      }),
      responseBody: "none",
      responseStatus: 500
    });
    onError(response);
  }
  async function fetchTransactionRequest(request, shouldClear) {
    if ("source" in request) {
      throw new Error(
        "Invalid request, transaction should be invoked only from a Frame. It was probably invoked from cast or composer action."
      );
    }
    const button = request.frameButton;
    const sourceFrame = request.sourceFrame;
    if (button.action !== "tx") {
      throw new Error("Invalid frame button action, tx expected");
    }
    if (request.signerStateActionContext.type !== "tx-data") {
      throw new Error(
        "Invalid signer state action context type, tx-data expected"
      );
    }
    if (shouldClear) {
      stackDispatch({ action: "CLEAR" });
    }
    const startTime = /* @__PURE__ */ new Date();
    const signedTransactionDataActionOrError = await tryCall(
      signFrameAction({
        actionContext: request.signerStateActionContext,
        // for transaction data we always use signer, so skip signing is false here
        forceRealSigner: true
      })
    );
    if (signedTransactionDataActionOrError instanceof Error) {
      onError(signedTransactionDataActionOrError);
      return;
    }
    const pendingItem = stackAPI.createPostPendingItem({
      action: signedTransactionDataActionOrError,
      request
    });
    signedTransactionDataActionOrError.searchParams.set(
      "specification",
      specification
    );
    const transactionDataResponse = await fetchProxied({
      proxyUrl: frameActionProxy,
      frameAction: signedTransactionDataActionOrError,
      fetchFn,
      specification,
      extraRequestPayload: extraButtonRequestPayload
    });
    const endTime = /* @__PURE__ */ new Date();
    if (transactionDataResponse instanceof Error) {
      stackAPI.markAsFailed({
        endTime,
        pendingItem,
        requestError: transactionDataResponse,
        response: null,
        responseBody: "none",
        responseStatus: 500
      });
      onError(transactionDataResponse);
      return;
    }
    if (!transactionDataResponse.ok) {
      await handleFailedResponse({
        response: transactionDataResponse,
        endTime,
        frameStackPendingItem: pendingItem
      });
      return;
    }
    function isTransactionTargetResponse(response) {
      return typeof response === "object" && response !== null && "method" in response;
    }
    const transactionData = await transactionDataResponse.json();
    if (!isTransactionTargetResponse(transactionData)) {
      const error = new Error(
        "The server returned an unexpected response for transaction data."
      );
      stackAPI.markAsFailed({
        endTime,
        pendingItem,
        requestError: error,
        response: transactionDataResponse,
        responseBody: transactionData,
        responseStatus: 500
      });
      onError(error);
      return;
    }
    let transactionIdOrError;
    if (transactionData.method === "eth_sendTransaction") {
      if (transactionData.params.data && transactionData.attribution !== false && transactionDataSuffix && // Has a function signature
      hexToBytes(transactionData.params.data).length > 4) {
        transactionData.params.data = transactionData.params.data + transactionDataSuffix.slice(2);
      }
      transactionIdOrError = await tryCall(
        onTransaction({
          frame: sourceFrame,
          frameButton: request.frameButton,
          transactionData
        }).then((transactionId) => {
          if (!transactionId) {
            return new Error("onTransaction did not return transaction id");
          }
          return transactionId;
        })
      );
    } else {
      transactionIdOrError = await tryCall(
        onSignature({
          frame: sourceFrame,
          frameButton: request.frameButton,
          signatureData: transactionData
        }).then((signatureHash) => {
          if (!signatureHash) {
            return new Error("onSignature did not return signature");
          }
          return signatureHash;
        })
      );
    }
    if (transactionIdOrError instanceof Error) {
      stackAPI.markAsFailed({
        pendingItem,
        endTime,
        requestError: transactionIdOrError,
        response: null,
        responseBody: "none",
        responseStatus: 500
      });
      onError(transactionIdOrError);
      return;
    }
    await fetchPOSTRequest(
      {
        ...request,
        signerStateActionContext: {
          ...request.signerStateActionContext,
          type: "tx-post",
          // include transactionId in payload
          transactionId: transactionIdOrError,
          // override target so the the request is sent to proper endpoint
          target: button.post_url || sourceFrame.postUrl || button.target
        }
      },
      // we are continuing with the same pending item
      {
        pendingFrameStackItem: pendingItem,
        startTime
      }
    );
  }
  async function fetchCastActionRequest(request, shouldClear = false) {
    const frameButton = {
      action: "post",
      label: request.action.name,
      target: request.action.action.postUrl || request.action.url
    };
    const signerStateActionContext = {
      ...request.signerStateActionContext,
      type: "default",
      frameButton
    };
    const signedDataOrError = await signAndGetFrameActionBodyPayload({
      signerStateActionContext,
      signFrameAction
    });
    if (shouldClear) {
      stackAPI.clear();
    }
    if (signedDataOrError instanceof Error) {
      onError(signedDataOrError);
      throw signedDataOrError;
    }
    const pendingItem = stackAPI.createCastOrComposerActionPendingItem({
      action: signedDataOrError,
      request: {
        ...request,
        frameButton,
        signerStateActionContext,
        method: "POST",
        source: "cast-action",
        sourceFrame: void 0
      }
    });
    const actionResponseOrError = await fetchProxied({
      fetchFn,
      proxyUrl: frameActionProxy,
      specification,
      frameAction: signedDataOrError,
      extraRequestPayload: extraButtonRequestPayload
    });
    if (actionResponseOrError instanceof Error) {
      onError(actionResponseOrError);
      throw actionResponseOrError;
    }
    try {
      const endTime = /* @__PURE__ */ new Date();
      if (!actionResponseOrError.ok) {
        await handleFailedResponse({
          response: actionResponseOrError,
          endTime,
          frameStackPendingItem: pendingItem
        });
        return;
      }
      const actionResponse = await actionResponseOrError.clone().json();
      if (isCastMessageResponse(actionResponse)) {
        stackAPI.markCastMessageAsDone({
          pendingItem,
          endTime,
          response: actionResponseOrError,
          responseData: actionResponse
        });
        return;
      }
      if (isCastActionFrameResponse(actionResponse)) {
        stackAPI.markCastFrameAsDone({ pendingItem, endTime });
        await fetchPOSTRequest({
          sourceFrame: void 0,
          frameButton: {
            action: "post",
            label: "action",
            target: actionResponse.frameUrl
          },
          isDangerousSkipSigning: request.isDangerousSkipSigning,
          method: "POST",
          source: "cast-action",
          signerStateActionContext: {
            ...request.signerStateActionContext,
            type: "default",
            buttonIndex: 1,
            frameButton: {
              action: "post",
              label: "action",
              target: actionResponse.frameUrl
            },
            target: actionResponse.frameUrl
          }
        });
        return;
      }
      throw new UnexpectedCastActionResponseError();
    } catch (e) {
      let error;
      if (!(e instanceof UnexpectedCastActionResponseError)) {
        console.error(`Unexpected response from the server`, e);
        error = e instanceof Error ? e : new Error("Unexpected error");
      } else {
        error = e;
      }
      onError(error);
      throw error;
    }
  }
  async function fetchComposerActionRequest(request, shouldClear = false) {
    const frameButton = {
      action: "post",
      label: request.action.name,
      target: request.action.url
    };
    const signerStateActionContext = {
      ...request.signerStateActionContext,
      type: "default",
      frameButton,
      state: encodeURIComponent(
        JSON.stringify({
          cast: request.composerActionState
        })
      )
    };
    const signedDataOrError = await signAndGetFrameActionBodyPayload({
      signerStateActionContext,
      signFrameAction
    });
    if (shouldClear) {
      stackAPI.clear();
    }
    if (signedDataOrError instanceof Error) {
      onError(signedDataOrError);
      throw signedDataOrError;
    }
    const pendingItem = stackAPI.createCastOrComposerActionPendingItem({
      action: signedDataOrError,
      request: {
        ...request,
        frameButton,
        signerStateActionContext,
        method: "POST",
        source: "composer-action",
        sourceFrame: void 0
      }
    });
    const actionResponseOrError = await fetchProxied({
      fetchFn,
      proxyUrl: frameActionProxy,
      specification,
      frameAction: signedDataOrError,
      extraRequestPayload: extraButtonRequestPayload
    });
    if (actionResponseOrError instanceof Error) {
      onError(actionResponseOrError);
      throw actionResponseOrError;
    }
    try {
      const endTime = /* @__PURE__ */ new Date();
      if (!actionResponseOrError.ok) {
        await handleFailedResponse({
          response: actionResponseOrError,
          endTime,
          frameStackPendingItem: pendingItem
        });
        return;
      }
      const actionResponse = await actionResponseOrError.clone().json();
      if (!isComposerFormActionResponse(actionResponse)) {
        throw new UnexpectedComposerActionResponseError();
      }
      stackAPI.markComposerFormActionAsDone({ pendingItem, endTime });
      await onComposerFormAction({
        form: actionResponse,
        cast: {
          embeds: [],
          text: "Cast text"
        }
      });
    } catch (e) {
      let error;
      if (!(e instanceof UnexpectedComposerActionResponseError)) {
        console.error(`Unexpected response from the server`, e);
        error = e instanceof Error ? e : new Error("Unexpected error");
      } else {
        error = e;
      }
      onError(error);
      throw error;
    }
  }
  return (request, shouldClear = false) => {
    if (request.method === "GET") {
      return fetchGETRequest(request, shouldClear);
    }
    if (request.method === "CAST_ACTION") {
      return fetchCastActionRequest(request, shouldClear);
    }
    if (request.method === "COMPOSER_ACTION") {
      return fetchComposerActionRequest(request, shouldClear);
    }
    if (request.frameButton.action === "tx") {
      return fetchTransactionRequest(request, shouldClear);
    }
    return fetchPOSTRequest(request, void 0, shouldClear);
  };
}
function proxyUrlAndSearchParamsToUrl(proxyUrl, ...searchParams) {
  const temporaryDomain = "temporary-for-parsing-purposes.tld";
  const parsedProxyUrl = new URL(proxyUrl, `http://${temporaryDomain}`);
  searchParams.forEach((params) => {
    params.forEach((value, key) => {
      parsedProxyUrl.searchParams.set(key, value);
    });
  });
  return parsedProxyUrl.hostname === temporaryDomain ? `${parsedProxyUrl.pathname}${parsedProxyUrl.search}` : parsedProxyUrl.toString();
}
async function fetchProxied(params) {
  const searchParams = new URLSearchParams({
    specification: params.specification
  });
  if ("frameAction" in params) {
    const proxyUrl2 = proxyUrlAndSearchParamsToUrl(
      params.proxyUrl,
      searchParams,
      params.frameAction.searchParams
    );
    return tryCall(
      params.fetchFn(proxyUrl2, {
        method: "POST",
        headers: {
          "Content-Type": "application/json"
        },
        body: JSON.stringify({
          ...params.extraRequestPayload,
          ...params.frameAction.body
        })
      })
    );
  }
  searchParams.set("url", params.url);
  const proxyUrl = proxyUrlAndSearchParamsToUrl(params.proxyUrl, searchParams);
  return tryCall(params.fetchFn(proxyUrl, { method: "GET" }));
}
async function tryCall(promise) {
  return promise.catch((e) => {
    if (e instanceof Error) {
      return e;
    }
    console.error(e);
    return new TypeError("Unexpected error, check the console for details");
  });
}
function getResponseBody(response) {
  if (response.headers.get("content-type")?.includes("/json")) {
    return response.clone().json();
  }
  return response.clone().text();
}
async function signAndGetFrameActionBodyPayload({
  signerStateActionContext,
  signFrameAction
}) {
  const { address: _, ...requiredFrameContext } = signerStateActionContext.frameContext;
  return tryCall(
    signFrameAction({
      actionContext: {
        ...signerStateActionContext,
        frameContext: requiredFrameContext
      }
    })
  );
}

export {
  useFetchFrame
};
